﻿/****************************************************************************

 ■HiResAmp 全体制御(ControlApp)
		Copyright(C) 2020 電子うさぎ https://電子うさぎ.com/

****************************************************************************/

using System;
using System.IO;
using System.Threading;
using System.Windows.Forms;
using System.Drawing;
using System.Runtime.InteropServices;
using System.Configuration;

namespace ControlApp
{
    public partial class MainForm : Form
    {
        private readonly string[] StateNames = { "",
            "停止",
            "開始中",
            "プレイヤー起動完了待ち",
            "マイミュージックへ移動",
            "マイミュージックへ移動待ち",
            "プレイリストへ移動",
            "プレイリストへ移動待ち",
            "個別プレイリストへ移動",
            "個別プレイリストへ移動待ち",
            "再生開始",
            "再生開始待ち",
            "再生終了待ち",
            "プレイリスト変更",
            "ウェイト"
        };

        private string m_SetFile;
        private Player m_Player;
        private Point m_LastPos;

        //---------------------------------------------------------------

        /// <summary>
        /// ●コンストラクタ
        /// </summary>
        public MainForm()
        {
            InitializeComponent();

            try
            {
                string path = System.AppDomain.CurrentDomain.BaseDirectory;

                if (!File.Exists(Path.Combine(path, "ControlApp.exe.config")))
                    throw new Exception("ControlApp.exe.config ファイルがありません");

                if (!File.Exists(Path.Combine(path, "Settings.txt")))
                    throw new Exception("Settings.txt ファイルがありません");

                m_SetFile = Path.Combine(path, "Settings.txt");

                m_Player = new Player(OnStateChange, OnPlayerException);
            }
            catch (Exception ex)
            {
                WindowState = FormWindowState.Normal;
                ShowException(ex);
                Close();
            }
        }

        /// <summary>
        /// ●ロード
        /// </summary>
        private void MainForm_Load(object sender, EventArgs e)
        {
            m_TopMost.Checked = TopMost;

            // テキストボックスのタブサイズを設定
            int[] arg = new int[] { 4 * 4 };
            IntPtr ptr = Marshal.AllocCoTaskMem(Marshal.SizeOf(typeof(int)) * arg.Length);
            Marshal.Copy(arg, 0, ptr, arg.Length);
            WinApi.SendMessage(m_Settings.Handle, 203, 1, ptr.ToInt32());
            Marshal.FreeCoTaskMem(ptr);

            m_SerialPort.NewLine = "\r\n";
            m_PortName.Text = ConfigurationManager.AppSettings["PortName"];

            try
            {
                // 初期化
                Init();

                // 起動時の自動開始
                if (m_Player.IsEnable)
                {
                    string[] args = Environment.GetCommandLineArgs();
                    if (args.Length < 2 || args[1] != "/D")
                        m_Player.Start();
                }
            }
            catch (Exception ex)
            {
                m_Player.Stop();

                if (m_SerialPort.IsOpen)
                {
                    try { m_SerialPort.WriteLine("ERR"); }
                    catch { }
                }

                WindowState = FormWindowState.Normal;
                ShowException(ex);
            }
        }

        /// <summary>
        /// ●再初期化
        /// </summary>
        private void m_Reset_Click(object sender, EventArgs e)
        {
            try
            {
                // 初期化
                Init();
            }
            catch (Exception ex)
            {
                m_Player.Stop();

                if (m_SerialPort.IsOpen)
                {
                    try { m_SerialPort.WriteLine("ERR"); }
                    catch { }
                }

                ShowException(ex);
            }
        }

        /// <summary>
        /// ●初期化
        /// </summary>
        private void Init()
        {
            m_Start.Enabled = false;

            m_Player.Stop();

            if (m_SerialPort.IsOpen)
                m_SerialPort.Close();

            if (m_Settings.Text == "")
            {
                using (StreamReader sr = new StreamReader(m_SetFile))
                {
                    m_Settings.Text = sr.ReadToEnd();
                    m_Settings.SelectionStart = 0;
                }
            }

            m_SerialPort.PortName = m_PortName.Text;
            m_SerialPort.Open();

            m_Player.Setting(m_Settings.Text);

            m_SerialPort.WriteLine("STARTED");
            m_Start.Enabled = m_Player.IsEnable;
        }

        /// <summary>
        /// ●設定保存
        /// </summary>
        private void m_Apply_Click(object sender, EventArgs e)
        {
            // 保存
            try
            {
                if (m_Settings.Text != string.Empty)
                {
                    using (StreamWriter sw = File.CreateText(m_SetFile))
                    {
                        sw.Write(m_Settings.Text);
                    }
                }
                Configuration config = ConfigurationManager.OpenExeConfiguration(ConfigurationUserLevel.None);
                config.AppSettings.Settings["PortName"].Value = m_PortName.Text;
                config.Save();
            }
            catch (Exception ex)
            {
                ShowException(ex);
                return;
            }
        }

        /// <summary>
        /// ●シリアル受信
        /// </summary>
        private void m_SerialPort_DataReceived(object sender, System.IO.Ports.SerialDataReceivedEventArgs e)
        {
            OnReceive(m_SerialPort.ReadLine());
        }

        /// <summary>
        /// ●シリアル受信
        /// </summary>
        private void OnReceive(string text)
        {
            if (InvokeRequired)
            {
                BeginInvoke(new Action<string>(OnReceive), new object[] { text });
                return;
            }
            try
            {
                switch (text)
                {
                case "PREV1":
                    m_SerialPort.WriteLine("OK");
                    m_Player.PrevOrder();
                    break;

                case "PREV2":
                    m_SerialPort.WriteLine("OK");
                    m_Player.PrevGenre();
                    break;

                case "NEXT1":
                    m_SerialPort.WriteLine("OK");
                    m_Player.NextOrder();
                    break;

                case "NEXT2":
                    m_SerialPort.WriteLine("OK");
                    m_Player.NextGenre();
                    break;

                case "SHUTDOWN":
                    m_SerialPort.WriteLine("OK");
                    m_Player.Stop();

                    Thread.Sleep(100);
                    m_SerialPort.Close();

                    // シャットダウン
                    WinApi.AdjustToken();
                    WinApi.ExitWindowsEx(WinApi.ExitWindows.EWX_POWEROFF | WinApi.ExitWindows.EWX_FORCE, 0);

                    Close();
                    break;
                }
            }
            catch (Exception ex)
            {
                m_Player.Stop();

                try { m_SerialPort.WriteLine("ERR"); }
                catch { }

                ShowException(ex);
            }
        }

        /// <summary>
        /// ●プレイヤー状態変化
        /// </summary>
        private void OnStateChange(Player.ProcState state)
        {
            m_State.Text = StateNames[(int)state];

            m_PrevGenre.Enabled = (state == Player.ProcState.WaitEnd);
            m_NextGenre.Enabled = m_PrevGenre.Enabled;
            m_PrevOrder.Enabled = m_PrevGenre.Enabled;
            m_NextOrder.Enabled = m_PrevGenre.Enabled;
        }

        /// <summary>
        /// ●プレイヤー例外
        /// </summary>
        private void OnPlayerException(Exception ex)
        {
            try { m_SerialPort.WriteLine("ERR"); }
            catch { }

            ShowException(ex);
        }

        /// <summary>
        /// ●クリック位置
        /// </summary>
        private void m_Click_Click(object sender, EventArgs e)
        {
            if (!m_Click.Checked)
            {
                MouseHook.StartHook(OnMouseHook);
                m_Click.Checked = true;
                m_Pixel.Checked = false;
            }
            else
            {
                StopMouseSetting();
            }
        }

        /// <summary>
        /// ●判断ピクセル
        /// </summary>
        private void m_Pixel_Click(object sender, EventArgs e)
        {
            if (!m_Pixel.Checked)
            {
                MouseHook.StartHook(OnMouseHook);
                m_Pixel.Checked = true;
                m_Click.Checked = false;
            }
            else
            {
                StopMouseSetting();
            }
        }

        /// <summary>
        /// ●開始/停止
        /// </summary>
        private void m_Stop_Click(object sender, EventArgs e)
        {
            if (m_Player.State == Player.ProcState.Stopped)
            {
                m_Player.Start();
                m_SerialPort.WriteLine("STARTED");
            }
            else
            {
                m_Player.Stop();
            }
        }

        /// <summary>
        /// ●マウスフックハンドラ
        /// </summary>
        private bool OnMouseHook(uint wParam, Point pt)
        {
            try
            {
                if (pt == m_LastPos)
                    return true;

                m_LastPos = pt;

                // カーソル周辺画像取得
                using (Bitmap src = new Bitmap(11, 11))
                {
                    using (Graphics g = Graphics.FromImage(src))
                    {
                        Point org = pt;
                        org.Offset(-5, -5);
                        g.CopyFromScreen(org, Point.Empty, src.Size);
                    }

                    // カーソル周辺拡大画像表示
                    using (Bitmap bmp = new Bitmap(110, 110))
                    {
                        using (Graphics g = Graphics.FromImage(bmp))
                        {
                            for (int x = 0; x < 11; x++)
                            {
                                for (int y = 0; y < 11; y++)
                                {
                                    SolidBrush brs = new SolidBrush(src.GetPixel(x, y));
                                    Rectangle rect = new Rectangle(x * 10, y * 10, 10, 10);
                                    g.FillRectangle(brs, rect);
                                }
                            }
                            g.DrawRectangle(new Pen(Color.Red), new Rectangle(50, 50, 9, 9));

                            m_Zoom.Image = bmp;
                            m_Zoom.Update();
                        }
                    }
                }

                // 座標表示
                pt = m_Player.ToClient(pt);
                m_Cood.Text = "X:" + pt.X + " Y:" + pt.Y;
            }
            catch(Exception ex)
            {
                StopMouseSetting();
                ShowException(ex);
            }

            return true;
        }

        /// <summary>
        /// ●ダイアログキー処理
        /// </summary>
        protected override bool ProcessDialogKey(Keys keyData)
        {
            if (m_Click.Checked || m_Pixel.Checked)
            {
                if (keyData == Keys.Left || keyData == Keys.Up || keyData == Keys.Right || keyData == Keys.Down)
                {
                    AdjustCursor(keyData);
                    return true;
                }
            }
            return base.ProcessDialogKey(keyData);
        }

        /// <summary>
        /// ●キー押下
        /// </summary>
        private void MainForm_KeyDown(object sender, KeyEventArgs e)
        {
            if (!m_Click.Checked && !m_Pixel.Checked)
                return;

            if (e.KeyCode == Keys.Escape)
            {
                if (m_Click.Checked)
                {
                    // クリック位置
                    Point pt = m_Player.ToClient(Cursor.Position);
                    m_Settings.SelectedText = "X=" + pt.X + ",Y=" + pt.Y;
                }
                else
                {
                    // 判断ピクセル
                    using (Bitmap bmp = new Bitmap(1, 1))
                    {
                        using (Graphics g = Graphics.FromImage(bmp))
                        {
                            Point pt = Cursor.Position;
                            g.CopyFromScreen(pt, Point.Empty, bmp.Size);
                            Color col = bmp.GetPixel(0, 0);

                            pt = m_Player.ToClient(pt);
                            m_Settings.SelectedText = "X=" + pt.X + ",Y=" + pt.Y + ",#"
                                 + col.R.ToString("X2") + col.G.ToString("X2") + col.B.ToString("X2");
                        }
                    }
                }
                e.SuppressKeyPress = true;
            }
            else if (e.KeyCode == Keys.Left || e.KeyCode == Keys.Up || e.KeyCode == Keys.Right || e.KeyCode == Keys.Down)
            {
                AdjustCursor(e.KeyCode);
                e.SuppressKeyPress = true;
            }
        }

        /// <summary>
        /// ●マウスカーソル微調整
        /// </summary>
        private void AdjustCursor(Keys keyData)
        {
            Point pt = Cursor.Position;
            switch(keyData)
            {
            case Keys.Left: --pt.X; break;
            case Keys.Right: ++pt.X; break;
            case Keys.Up: --pt.Y; break;
            case Keys.Down: ++pt.Y; break;
            }
            Cursor.Position = pt;

            OnMouseHook(0, pt);
        }

        /// <summary>
        /// ●前のジャンル
        /// </summary>
        private void m_PrevGenre_Click(object sender, EventArgs e)
        {
            m_Player.PrevGenre();
        }

        /// <summary>
        /// ●次のジャンル
        /// </summary>
        private void m_NextGenre_Click(object sender, EventArgs e)
        {
            m_Player.NextGenre();
        }

        /// <summary>
        /// ●前のプレイリスト
        /// </summary>
        private void m_PrevOrder_Click(object sender, EventArgs e)
        {
            m_Player.PrevOrder();
        }

        /// <summary>
        /// ●次のプレイリスト
        /// </summary>
        private void m_NextOrder_Click(object sender, EventArgs e)
        {
            m_Player.NextOrder();
        }

        /// <summary>
        /// ●TopMost
        /// </summary>
        private void m_TopMost_CheckedChanged(object sender, EventArgs e)
        {
            TopMost = m_TopMost.Checked;
        }

        /// <summary>
        /// ●カーソル移動
        /// </summary>
        private void m_MoveCursor_CheckedChanged(object sender, EventArgs e)
        {
            m_Player.CursorMove = m_CursorMove.Checked;
        }

        /// <summary>
        /// ●例外メッセージを表示する
        /// </summary>
        private void ShowException(Exception ex)
        {
            MessageBox.Show(this, ex.Message + "\r\n\r\n" + ex.StackTrace,
                this.Text, MessageBoxButtons.OK, MessageBoxIcon.Exclamation);
        }

        /// <summary>
        /// ●フォーム最小化
        /// </summary>
        private void MainForm_SizeChanged(object sender, EventArgs e)
        {
            if (WindowState == FormWindowState.Minimized)
                StopMouseSetting();
        }

        /// <summary>
        /// ●フォームを閉じる
        /// </summary>
        private void MainForm_FormClosing(object sender, FormClosingEventArgs e)
        {
            MouseHook.EndHook();
        }

        /// <summary>
        /// ●マウス設定を停止する
        /// </summary>
        private void StopMouseSetting()
        {
            MouseHook.EndHook();
            m_Click.Checked = false;
            m_Pixel.Checked = false;
            m_Zoom.Image = null;
            m_Cood.Text = string.Empty;
            m_LastPos.X = int.MinValue;
            m_LastPos.Y = int.MinValue;
        }
    }
}
