﻿/****************************************************************************

 ■HiResAmp 全体制御(ControlApp)
		Copyright(C) 2020 電子うさぎ https://電子うさぎ.com/

****************************************************************************/

using System;
using System.Drawing;
using System.Diagnostics;
using System.Windows.Forms;
using System.Collections.Generic;
using System.Text.RegularExpressions;
using System.Configuration;

namespace ControlApp
{
    class Player
    {
        public enum ProcState
        {
            None,
            Stopped,        // 停止
            Started,        // 開始中
            WaitPlayer,     // プレイヤー起動完了待ち
            ToMyMusic,      // マイミュージックへ移動
            WaitToMyMusic,  // マイミュージックへ移動待ち
            ToPlayList,     // プレイリストへ移動
            WaitToPlayList, // プレイリストへ移動待ち
            ToPlayEntry,    // 個別プレイリストへ移動
            WaitToPlayEntry,// 個別プレイリストへ移動待ち
            StartPlay,      // 再生開始       
            WaitStartPlay,  // 再生開始ウェイト
            WaitEnd,        // 再生終了待ち
            ChangePlayList, // プレイリスト変更
            Wait,           // ウェイト
        }

        public enum FindType
        {
            Normal,         // 標準(NextOrder)
            PrevOrder,      // 現在のジャンルの前のプレイリスト
            NextOrder,      // 現在のジャンルの次のプレイリスト
            PrevGenre,      // 前のジャンルの最初のプレイリスト
            NextGenre,      // 次のジャンルの最初のプレイリスト
        }

        public ProcState State { get; set; }
        public bool CursorMove { get; set; }
        public bool IsEnable { get { return m_Setting; } }

        //---------------------------------------------------------------

        private const int TimerInterval = 250;

        private class PosColor
        {
            public Point Pt { get; set; }
            public Color Col { get; set; }
        }
        private class PlayList
        {
            public int Order { get; set; }
            public int Genre { get; set; }
        }

        private Timer m_Timer;
        private Action<ProcState> m_OnStateChange;
        private Action<Exception> m_OnException;
        private IntPtr m_AmazonMusic;
        private IntPtr m_OpeWnd;
        private bool m_Setting;

        private Regex m_RegPos;
        private Regex m_RegCol;
        private Regex m_RegVal;
        private Regex m_RegPl;

        private Dictionary<string, PosColor> m_PosCols;
        private Dictionary<string, int> m_Values;
        private List<PlayList> m_PlayList;

        private int m_Counter;
        private int m_WaitTime;
        private ProcState m_Next;

        private int m_CurOrder;
        private int m_CurGenre;
        private int m_NextOrder;
        private FindType m_FindType;
        private int m_ErrCount;

        //---------------------------------------------------------------

        /// <summary>
        /// ●コンストラクタ
        /// </summary>
        internal Player(Action<ProcState> sc, Action<Exception> ex)
        {
            State = ProcState.Stopped;
            CursorMove = false;
            m_Setting = false;
            m_AmazonMusic = IntPtr.Zero;
            m_OpeWnd = IntPtr.Zero;

            m_Timer = new Timer();
            m_Timer.Interval = TimerInterval;
            m_Timer.Tick += OnTimer;

            m_OnStateChange = sc;
            m_OnException = ex;

            m_RegPos = new Regex(@"^(\w.+):\s*X\s*=\s*(\d+)\s*\,\s*Y\s*=\s*(\d+)\s*\,");
            m_RegCol = new Regex(@"^(\w.+):\s*X\s*=\s*(\d+)\s*\,\s*Y\s*=\s*(\d+)\s*\,#([a-zA-Z0-9]{2})([a-zA-Z0-9]{2})([a-zA-Z0-9]{2})\,");
            m_RegVal = new Regex(@"^(\w.+):\s*(\d+)\s*\,");
            m_RegPl = new Regex(@"^(\d{2})-G(\d{1}):\s*(\d{1})");

            m_PosCols = new Dictionary<string, PosColor>();
            m_Values = new Dictionary<string, int>();
            m_PlayList = new List<PlayList>();

            int.TryParse(ConfigurationManager.AppSettings["LastOrder"], out m_CurOrder);
            int.TryParse(ConfigurationManager.AppSettings["LastGenre"], out m_CurGenre);
            m_NextOrder = 0;
            m_FindType = FindType.Normal;
        }

        /// <summary>
        /// ●設定
        /// </summary>
        internal void Setting(string text)
        {
            m_Setting = false;
            m_PosCols.Clear();
            m_Values.Clear();
            m_PlayList.Clear();

            string[] lines = text.Split(new string[] { "\r\n" }, StringSplitOptions.RemoveEmptyEntries);

            foreach (string line in lines)
            {
                // コメント
                if (line.StartsWith("#") || line.StartsWith("●"))
                    continue;

                // 共通設定(クリック位置)
                Match m = m_RegCol.Match(line);
                if (m.Success)
                {
                    PosColor pc = new PosColor();
                    pc.Pt = new Point(int.Parse(m.Groups[2].Value), int.Parse(m.Groups[3].Value));
                    pc.Col = Color.FromArgb(Convert.ToInt32(m.Groups[4].Value, 16),
                        Convert.ToInt32(m.Groups[5].Value, 16), Convert.ToInt32(m.Groups[6].Value, 16));

                    m_PosCols.Add(m.Groups[1].Value, pc);
                    continue;
                }

                // 共通設定(判断ピクセル)
                m = m_RegPos.Match(line);
                if (m.Success)
                {
                    PosColor pc = new PosColor();
                    pc.Pt = new Point(int.Parse(m.Groups[2].Value), int.Parse(m.Groups[3].Value));
                    pc.Col = Color.White;

                    m_PosCols.Add(m.Groups[1].Value, pc);
                    continue;
                }

                // 共通設定(値)
                m = m_RegVal.Match(line);
                if (m.Success)
                {
                    m_Values.Add(m.Groups[1].Value, int.Parse(m.Groups[2].Value));
                    continue;
                }

                // プレイリスト有効無効設定
                m = m_RegPl.Match(line);
                if (m.Success)
                {
                    if (m.Groups[3].Value != "0")
                    {
                        PlayList gen = new PlayList();
                        gen.Order = int.Parse(m.Groups[1].Value);
                        gen.Genre = int.Parse(m.Groups[2].Value);
                        m_PlayList.Add(gen);
                    }
                    continue;
                }

                throw new Exception("不正な設定項目です：" + line);
            }

            // チェック
            string[] keys1 = {
                "WaitPlayerPixel", "MyMusicClick", "MyMusicPixel", "PlayListClick", "PlayListPixel",
                "PlayEntryClick",  "PlayEntryMusic", "PlayEntryPixel", "PlayClick", "PlayStartPixel", "PlayEndPixel" };
            foreach(string k in keys1)
            {
                if (!m_PosCols.ContainsKey(k))
                    throw new Exception("設定項目がありません：" + k);
            }

            string[] keys2 = { "PlayEntryWidth", "PlayEntryHeight", "PlayEntryColNum", "StartTimeout",
                "TransitionWait", "BeforePlayWait", "EndJudgTime" };
            foreach (string k in keys2)
            {
                if (!m_Values.ContainsKey(k))
                    throw new Exception("設定項目がありません：" + k);
            }

            if (m_PlayList.Count == 0)
                throw new Exception("有効なプレイリストがありません：");

            m_Setting = true;
        }

        /// <summary>
        /// ●開始
        /// </summary>
        internal void Start()
        {
            m_Timer.Stop();
            m_Counter = 0;
            m_ErrCount = 0;

            SetState(ProcState.Started);
            m_Timer.Start();
        }

        /// <summary>
        /// ●停止
        /// </summary>
        internal void Stop()
        {
            if (State == ProcState.WaitEnd)
            {
                if (m_OpeWnd != IntPtr.Zero)
                    DoClick("StopClick");
            }
            m_Timer.Stop();
            SetState(ProcState.Stopped);
        }

        /// <summary>
        /// ●前のプレイリスト
        /// </summary>
        internal void PrevOrder()
        {
            if (State == ProcState.WaitStartPlay || State == ProcState.WaitEnd)
            {
                m_FindType = FindType.PrevOrder;
                SetState(ProcState.ChangePlayList);
            }
        }

        /// <summary>
        /// ●次のプレイリスト
        /// </summary>
        internal void NextOrder()
        {
            if (State == ProcState.WaitStartPlay || State == ProcState.WaitEnd)
            {
                m_FindType = FindType.NextOrder;
                SetState(ProcState.ChangePlayList);
            }
        }

        /// <summary>
        /// ●前のジャンル
        /// </summary>
        internal void PrevGenre()
        {
            if (State == ProcState.WaitStartPlay || State == ProcState.WaitEnd)
            {
                m_FindType = FindType.PrevGenre;
                SetState(ProcState.ChangePlayList);
            }
        }

        /// <summary>
        /// ●次のジャンル
        /// </summary>
        internal void NextGenre()
        {
            if (State == ProcState.WaitStartPlay || State == ProcState.WaitEnd)
            {
                m_FindType = FindType.NextGenre;
                SetState(ProcState.ChangePlayList);
            }
        }

        /// <summary>
        /// ●スクリーン座標をクライアント座標へ変換する
        /// </summary>
        internal Point ToClient(Point pt)
        {
            WinApi.POINT temp = pt;
            WinApi.ScreenToClient(GetOpeWnd(), out temp);
            return temp;
        }

        //---------------------------------------------------------------

        /// <summary>
        /// ●メイン処理
        /// </summary>
        private void OnTimer(object sender, EventArgs e)
        {
            try
            {
                MainProc();
            }
            catch (Exception ex)
            {
                Stop();
                m_Timer.Stop();
                m_OnException(ex);
            }
        }

        /// <summary>
        /// ●メイン処理
        /// </summary>
        private bool MainProc()
        {
            switch (State)
            {
            case ProcState.Started:
                // AmazonMusicを探す
                if (FindAmazonMusic())
                {
                    WinApi.SetForegroundWindow(m_AmazonMusic);
                    SetState(ProcState.WaitPlayer);
                }
                else
                {
                    if (++m_Counter > m_Values["StartTimeout"] / TimerInterval)
                    {
                        if (m_AmazonMusic == IntPtr.Zero)
                            throw new Exception("Amazon Music が起動していません。");

                        throw new Exception("Amazon Music 操作ウィンドウが見つかりません。");
                    }
                }
                break;

            case ProcState.WaitPlayer:
                // プレイヤー起動完了待ち

                if (IsPosColor("WaitPlayerPixel"))
                {
                    m_NextOrder = GetNextOrder();
                    SetState(ProcState.ToMyMusic);
                }
                break;

            case ProcState.ToMyMusic:
                // マイミュージックへ移動する
                if (!IsPosColor("MyMusicPixel"))
                {
                    DoClick("MyMusicClick");
                    SetState(ProcState.WaitToMyMusic);
                }
                else
                {
                    SetState(ProcState.ToPlayList);
                }
                break;

            case ProcState.WaitToMyMusic:
                // マイミュージックへ移動待ち
                if (IsPosColor("MyMusicPixel"))
                {
                    Wait(ProcState.ToPlayList, m_Values["TransitionWait"]);
                }
                break;

            case ProcState.ToPlayList:
                // プレイリストへ移動する
                if (!IsPosColor("PlayListPixel"))
                {
                    DoClick("PlayListClick");

                    // ▶を出すためにマウスカーソルを移動
                    DoMove("PlayEntryClick");

                    SetState(ProcState.WaitToPlayList);
                }
                else
                {
                    SetState(ProcState.ToPlayEntry);
                }
                break;

            case ProcState.WaitToPlayList:
                // プレイリストへ移動待ち
                if (IsPosColor("PlayListPixel"))
                {
                    Wait(ProcState.ToPlayEntry, m_Values["TransitionWait"]);
                }
                break;

            case ProcState.ToPlayEntry:
                // 個別プレイリストへ移動する
                {
                    int x = (m_NextOrder - 1) % m_Values["PlayEntryColNum"];
                    int y = (m_NextOrder - 1) / m_Values["PlayEntryColNum"];
                    DoClick("PlayEntryClick", m_Values["PlayEntryWidth"] * x, m_Values["PlayEntryHeight"] * y);

                    // ▶を出すためにマウスカーソルを移動
                    DoMove("PlayEntryMusic");

                    SetState(ProcState.WaitToPlayEntry);
                }
                break;

            case ProcState.WaitToPlayEntry:
                // 個別プレイリストへ移動待ち
                if (IsPosColor("PlayEntryPixel"))
                {
                    Wait(ProcState.StartPlay, m_Values["BeforePlayWait"]);
                }
                break;

            case ProcState.StartPlay:
                // 再生開始
                DoClick("PlayClick");
                SetState(ProcState.WaitStartPlay);
                break;

            case ProcState.WaitStartPlay:
                // 再生開始待ち
                if (IsPosColor("PlayStartPixel"))
                {
                    SetState(ProcState.WaitEnd);
                }
                break;

            case ProcState.WaitEnd:
                // 再生終了待ち
                if (IsPosColor("PlayEndPixel"))
                {
                    // 終了したら次のプレイリストへ
                    if (++m_Counter > (m_Values["EndJudgTime"] / TimerInterval))
                        SetState(ProcState.ChangePlayList);
                }
                else
                {
                    m_Counter = 0;
                }
                break;

            case ProcState.ChangePlayList:
                {
                    // プレイリスト変更
                    //DoClick("StopClick"); ※Combo384の不具合?により次の再生の冒頭に前の曲が一瞬入るため止めない

                    int nNewOrder = GetNextOrder();
                    if (nNewOrder != m_NextOrder)
                    {
                        m_NextOrder = nNewOrder;
                        SetState(ProcState.ToMyMusic);
                    }
                    else
                    {
                        SetState(ProcState.StartPlay);
                    }
                }
                break;

            case ProcState.Wait:
                // ウェイト
                if (++m_Counter >= m_WaitTime / TimerInterval)
                {
                    SetState(m_Next);
                    m_Next = ProcState.None;
                }
                break;
            }

            return true;
        }

        /// <summary>
        /// ●AmazonMusicを探す
        /// </summary>
        private bool FindAmazonMusic()
        {
            m_AmazonMusic = IntPtr.Zero;
            m_OpeWnd = IntPtr.Zero;

            // Amazon Music メインウィンドウを探す
            m_AmazonMusic = WinApi.FindWindowEx(IntPtr.Zero, IntPtr.Zero, "Amazon Music", null);

            if (m_AmazonMusic == IntPtr.Zero)
                return false;

            // Amazon Music 操作ウィンドウを探す
            m_OpeWnd = WinApi.FindWindowEx(m_AmazonMusic, IntPtr.Zero, "CefBrowserWindow", null);
            if (m_OpeWnd == IntPtr.Zero)
                return false;

            m_OpeWnd = WinApi.FindWindowEx(m_OpeWnd, IntPtr.Zero, "Chrome_WidgetWin_0", null);
            if (m_OpeWnd == IntPtr.Zero)
                return false;

            m_OpeWnd = WinApi.FindWindowEx(m_OpeWnd, IntPtr.Zero, null, "Chrome Legacy Window");
            if (m_OpeWnd == IntPtr.Zero)
                return false;

            return true;
        }

        /// <summary>
        /// ●状態を設定する
        /// </summary>
        private void SetState(ProcState state)
        {
            State = state;
            m_OnStateChange(state);
        }

        /// <summary>
        /// ●指定したクライアント位置が指定した色かどうかを調べる
        /// </summary>
        private bool IsPosColor(string key)
        {
            PosColor pc = m_PosCols[key];
            WinApi.POINT pt = pc.Pt;
            WinApi.ClientToScreen(GetOpeWnd(), out pt);

            // リモートデスクトップで接続後にハンドル無効例外が出るためリトライする
            for (int i = 0; ; i++)
            {
                using (Bitmap bmp = new Bitmap(1, 1))
                {
                    try
                    {
                        using (Graphics g = Graphics.FromImage(bmp))
                        {
                            //g.CopyFromScreen(pt, Point.Empty, bmp.Size); // リモートデスクトップで例外が出るのはコレのせいだった

                            IntPtr hDstDC = IntPtr.Zero;
                            IntPtr hSrcDC = IntPtr.Zero;
                            try
                            {
                                hDstDC = g.GetHdc();
                                hSrcDC = WinApi.GetDC(IntPtr.Zero); //WinApi.GetWindowDC(GetOpeWnd());

                                WinApi.BitBlt(hDstDC, 0, 0, bmp.Width, bmp.Height, hSrcDC, pt.X, pt.Y, WinApi.SRCCOPY);

                                g.ReleaseHdc(hDstDC);
                                hDstDC = IntPtr.Zero;

                                WinApi.ReleaseDC(IntPtr.Zero, hSrcDC);
                                hSrcDC = IntPtr.Zero;
                            }
                            catch
                            {
                                if (hDstDC != IntPtr.Zero)
                                {
                                    g.ReleaseHdc(hDstDC);
                                    hDstDC = IntPtr.Zero;
                                }
                                if (hSrcDC != IntPtr.Zero)
                                {
                                    WinApi.ReleaseDC(IntPtr.Zero, hSrcDC);
                                    hSrcDC = IntPtr.Zero;
                                }
                                throw;
                            }

                            Color col = bmp.GetPixel(0, 0);
                            return (col.R == pc.Col.R && col.G == pc.Col.G && col.B == pc.Col.B);
                        }
                    }
                    catch(Exception ex)
                    {
                        if (i >= 6)
                        {
                            if (++m_ErrCount > 10)
                                throw ex; // 30秒以上経ってもダメなら諦める

                            return false;
                        }
                        GC.Collect();
                        System.Threading.Thread.Sleep(500);
                    }
                }
            }
        }

        /// <summary>
        /// ●クリックする
        /// </summary>
        private void DoClick(string key)
        {
            DoClick(key, 0, 0);
        }

        /// <summary>
        /// ●クリックする
        /// </summary>
        private void DoClick(string key, int offsetX, int offsetY)
        {
            Point pt = m_PosCols[key].Pt;
            pt.X += offsetX;
            pt.Y += offsetY;
            DoClick(pt);
        }

        /// <summary>
        /// ●クリックする
        /// </summary>
        private void DoClick(Point pt)
        {
            if (CursorMove)
            {
                WinApi.POINT temp = pt;
                WinApi.ClientToScreen(GetOpeWnd(), out temp);
                Cursor.Position = temp;
            }

            int lParam = (pt.Y << 16 | pt.X);
            IntPtr hWnd = GetOpeWnd();
            WinApi.SendMessage(hWnd, WinApi.WM_LBUTTONDOWN, WinApi.MK_LBUTTON, lParam);
            WinApi.SendMessage(hWnd, WinApi.WM_LBUTTONUP, 0x00000000, lParam);
        }

        /// <summary>
        /// ●マウス移動
        /// </summary>
        private void DoMove(string key)
        {
            Point pt = m_PosCols[key].Pt;

            WinApi.POINT temp = pt;
            WinApi.ClientToScreen(GetOpeWnd(), out temp);
            Cursor.Position = temp;
        }

        /// <summary>
        /// ●ウェイトへ入る
        /// </summary>
        private void Wait(ProcState next, int ms)
        {
            m_Counter = 0;
            m_WaitTime = ms;
            m_Next = next;
            SetState(ProcState.Wait);
        }

        /// <summary>
        /// ●次のプレイリストオーダーを取得する
        /// </summary>
        private int GetNextOrder()
        {
            switch(m_FindType)
            {
            case FindType.Normal:
                m_CurOrder = FindNextOrder();
                break;

            case FindType.PrevOrder:    // 現在のジャンルの前のプレイリスト
                m_CurOrder = FindPrevOrder();
                break;
            case FindType.NextOrder:    // 現在のジャンルの次のプレイリスト
                m_CurOrder = FindNextOrder();
                break;
            case FindType.PrevGenre:    // 前のジャンルの最初のプレイリスト
                m_CurOrder = FindPrevGenre();
                break;
            case FindType.NextGenre:    // 次のジャンルの最初のプレイリスト
                m_CurOrder = FindNextGenre();
                break;
            }

            m_FindType = FindType.Normal;

            Configuration config = ConfigurationManager.OpenExeConfiguration(ConfigurationUserLevel.None);
            config.AppSettings.Settings["LastOrder"].Value = m_CurOrder.ToString();
            config.AppSettings.Settings["LastGenre"].Value = m_CurGenre.ToString();
            config.Save();

            return m_CurOrder;
        }

        /// <summary>
        /// ●現在のジャンルの前のプレイリストを取得する
        /// </summary>
        private int FindPrevOrder()
        {
            // プレイリストが存在しているかを確認する
            if (FixCurrent())
                return m_CurOrder;

            // 前のプレイリスト
            List<int> list = new List<int>();
            foreach (PlayList pl in m_PlayList)
            {
                if (pl.Genre == m_CurGenre)
                {
                    if (pl.Order != m_CurOrder)
                        list.Add(pl.Order);
                }
            }
            if (list.Count == 0)
                return m_CurOrder;

            if (list.Count == 1)
                return list[0];

            list.Sort();
            list.Reverse();
            foreach (int order in list)
            {
                if (order < m_CurOrder)
                    return order;
            }

            return list[0];
        }

        /// <summary>
        /// ●現在のジャンルの次のプレイリストを取得する
        /// </summary>
        private int FindNextOrder()
        {
            // プレイリストが存在しているかを確認する
            if (FixCurrent())
                return m_CurOrder;

            // 次のプレイリスト
            List<int> list = new List<int>();
            foreach (PlayList pl in m_PlayList)
            {
                if (pl.Genre == m_CurGenre)
                {
                    if (pl.Order != m_CurOrder)
                        list.Add(pl.Order);
                }
            }
            if (list.Count == 0)
                return m_CurOrder; // 無い場合は何もしない

            if (list.Count == 1)
                return list[0];

            list.Sort();
            foreach(int order in list)
            {
                if (order > m_CurOrder)
                    return order;
            }

            return list[0];
        }

        /// <summary>
        /// ●前のジャンルの最初のプレイリストを取得する
        /// </summary>
        private int FindPrevGenre()
        {
            // プレイリストが存在しているかを確認する
            if (FixCurrent())
                return m_CurOrder;

            // 次のジャンル
            int next = int.MinValue;
            foreach (PlayList pl in m_PlayList)
            {
                // 現在のジャンルより小さい最も大きなジャンル
                if (pl.Genre < m_CurGenre && pl.Genre > next)
                    next = pl.Genre;
            }
            if (next == int.MinValue)
            {
                // 無い場合最も大きなジャンル
                foreach (PlayList pl in m_PlayList)
                {
                    if (pl.Genre > next)
                        next = pl.Genre;
                }
            }
            if (next == int.MinValue)
                return m_CurOrder; // 無い場合は何もしない

            m_CurGenre = next;

            next = int.MaxValue;
            foreach (PlayList pl in m_PlayList)
            {
                // 最も小さなオーダー
                if (pl.Genre == m_CurGenre && pl.Order < next)
                    next = pl.Order;
            }

            return next;
        }

        /// <summary>
        /// ●次のジャンルの最初のプレイリストを取得する
        /// </summary>
        private int FindNextGenre()
        {
            // プレイリストが存在しているかを確認する
            if (FixCurrent())
                return m_CurOrder;

            // 次のジャンル
            int next = int.MaxValue;
            foreach (PlayList pl in m_PlayList)
            {
                // 現在のジャンルより大きい最も小さなジャンル
                if (pl.Genre > m_CurGenre && pl.Genre < next)
                    next = pl.Genre;
            }
            if (next == int.MaxValue)
            {
                // 無い場合最も小さなジャンル
                foreach (PlayList pl in m_PlayList)
                {
                    if (pl.Genre < next)
                        next = pl.Genre;
                }
            }
            if (next == int.MaxValue)
                return m_CurOrder; // 無い場合は何もしない

            m_CurGenre = next;

            next = int.MaxValue;
            foreach (PlayList pl in m_PlayList)
            {
                // 最も小さなオーダー
                if (pl.Genre == m_CurGenre && pl.Order < next)
                    next = pl.Order;
            }

            return next;
        }

        /// <summary>
        /// ●存在しているプレイリストへ変更する
        /// </summary>
        private bool FixCurrent()
        {
            // プレイリストが存在しているかを確認する
            foreach (PlayList pl in m_PlayList)
            {
                if (pl.Order == m_CurOrder && pl.Genre == m_CurGenre)
                    return false;
            }

            // 存在していない→最初のプレイリストに決定
            m_CurOrder = m_PlayList[0].Order;
            m_CurGenre = m_PlayList[0].Genre;

            return true;
        }

        /// <summary>
        /// ●操作ウィンドウハンドルを取得する
        /// </summary>
        private IntPtr GetOpeWnd()
        {
            try
            {
                // ハンドルの有効性チェック
                if (WinApi.IsWindow(m_OpeWnd))
                {
                    // ウィンドウを最大化する
                    if (!WinApi.IsZoomed(m_AmazonMusic))
                        WinApi.ShowWindow(m_AmazonMusic, WinApi.SW_MAXIMIZE);

                    return m_OpeWnd;
                }
            }
            catch { }

            // 取得し直す
            if (!FindAmazonMusic())
                throw new Exception("Amazon Music が見つかりません。");

            // ウィンドウを最大化する
            if (!WinApi.IsZoomed(m_AmazonMusic))
                WinApi.ShowWindow(m_AmazonMusic, WinApi.SW_MAXIMIZE);

            return m_OpeWnd;
        }
    }
}
