/****************************************************************************

 ■HiResAmp 入力監視(InputSection)
		Copyright(C) 2020 電子うさぎ https://電子うさぎ.com/

****************************************************************************/

// PIC32MX120F032B Configuration Bit Settings

// DEVCFG3
#pragma config USERID = 0xFFFF          // Enter Hexadecimal value (Enter Hexadecimal value)
#pragma config PMDL1WAY = OFF           // Peripheral Module Disable Configuration (Allow multiple reconfigurations)
#pragma config IOL1WAY = OFF            // Peripheral Pin Select Configuration (Allow multiple reconfigurations)

// DEVCFG2
#pragma config FPLLIDIV = DIV_2         // PLL Input Divider (2x Divider)
#pragma config FPLLMUL = MUL_24         // PLL Multiplier (24x Multiplier)
#pragma config FPLLODIV = DIV_2         // System PLL Output Clock Divider (PLL Divide by 2)

// DEVCFG1
#pragma config FNOSC = FRCDIV           // Oscillator Selection Bits (Fast RC Osc w/Div-by-N (FRCDIV))
#pragma config FSOSCEN = OFF            // Secondary Oscillator Enable (Disabled)
#pragma config IESO = ON                // Internal/External Switch Over (Enabled)
#pragma config POSCMOD = OFF            // Primary Oscillator Configuration (Primary osc disabled)
#pragma config OSCIOFNC = OFF           // CLKO Output Signal Active on the OSCO Pin (Disabled)
#pragma config FPBDIV = DIV_1           // Peripheral Clock Divisor (Pb_Clk is Sys_Clk/1)
#pragma config FCKSM = CSECMD           // Clock Switching and Monitor Selection (Clock Switch Enable, FSCM Disabled)
#pragma config WDTPS = PS8192           // Watchdog Timer Postscaler (1:8192)
#pragma config WINDIS = OFF             // Watchdog Timer Window Enable (Watchdog Timer is in Non-Window Mode)
#pragma config FWDTEN = OFF             // Watchdog Timer Enable (WDT Disabled (SWDTEN Bit Controls))
#pragma config FWDTWINSZ = WINSZ_25     // Watchdog Timer Window Size (Window Size is 25%)

// DEVCFG0
#pragma config JTAGEN = OFF             // JTAG Enable (JTAG Disabled)
#pragma config ICESEL = ICS_PGx1        // ICE/ICD Comm Channel Select (Communicate on PGEC1/PGED1)
#pragma config PWP = OFF                // Program Flash Write Protect (Disable)
#pragma config BWP = OFF                // Boot Flash Write Protect bit (Protection Disabled)
#pragma config CP = OFF                 // Code Protect (Protection Disabled)

#include <xc.h>
#include <proc/p32mx120f032b.h>
#include <GenericTypeDefs.h>

//=================================================================

#define I2C_SLAVE			0x30

#define DEBUG_LED			_LATA0
#define COMBO384_F3			_RB12
#define COMBO384_F2			_RB11
#define COMBO384_F1			_RB10
#define COMBO384_F0			_RB9
#define COMBO384_DSDON		_RB8
#define COMBO384_DSD64		_RB7
#define COMBO384_MUTE		_RB5
#define DCLK_IN				_RB15

#define COMBO384_F3_MASK	0x08
#define COMBO384_F2_MASK	0x04
#define COMBO384_F1_MASK	0x02
#define COMBO384_F0_MASK	0x01
#define COMBO384_DSDON_MASK	0x40
#define COMBO384_DSD64_MASK	0x20
#define COMBO384_MUTE_MASK	0x10

#define DSD_UNKNOWN			0
#define DSD_32KHZ			1		
#define DSD_44KHZ			2
#define DSD_48KHZ			3

#define DSD_FREQ_MARGIN		0.04		// ±4%の誤差を許容
#define MIN_DSD_FREQ(x)		(x * (1.0 - DSD_FREQ_MARGIN))
#define MAX_DSD_FREQ(x)		(x * (1.0 + DSD_FREQ_MARGIN))

#define CHECK_COUNT_MAX		512
#define CHECK_COUNT_THR		(CHECK_COUNT_MAX / 2)

//---------------------------------------------------------------

typedef struct {
	int nSam;
	UINT16 nMin;
	UINT16 nMax;
} SDsdSample;

const SDsdSample s_DsdSam[] = {
	{ DSD_32KHZ, MIN_DSD_FREQ(2048), MAX_DSD_FREQ(2048) },
	{ DSD_32KHZ, MIN_DSD_FREQ(4096), MAX_DSD_FREQ(4096) },
	{ DSD_32KHZ, MIN_DSD_FREQ(8192), MAX_DSD_FREQ(8192) },
	{ DSD_32KHZ, MIN_DSD_FREQ(16284), MAX_DSD_FREQ(16284) },
	{ DSD_44KHZ, MIN_DSD_FREQ(2822), MAX_DSD_FREQ(2822) },
	{ DSD_44KHZ, MIN_DSD_FREQ(5644), MAX_DSD_FREQ(5644) },
	{ DSD_44KHZ, MIN_DSD_FREQ(11289), MAX_DSD_FREQ(11289) },
	{ DSD_44KHZ, MIN_DSD_FREQ(22579), MAX_DSD_FREQ(22579) },
	{ DSD_48KHZ, MIN_DSD_FREQ(3072), MAX_DSD_FREQ(3072) },
	{ DSD_48KHZ, MIN_DSD_FREQ(6144), MAX_DSD_FREQ(6144) },
	{ DSD_48KHZ, MIN_DSD_FREQ(12288), MAX_DSD_FREQ(12288) },
	{ DSD_48KHZ, MIN_DSD_FREQ(24576), MAX_DSD_FREQ(24576) },
	{ DSD_UNKNOWN, 0, 0 }
};

static volatile UINT8 s_nDetected;

//---------------------------------------------------------------

static void InitSystem(void);
static void Run(void);
static void Clock(BOOL bFast);

#define EnableWdt()	WDTCONbits.ON = 1
#define ClrWdt()	WDTCONbits.WDTCLR = 1

//=================================================================

// ●メインエントリ
void main(void)
{
	InitSystem();
	Run();
}

// ●システム初期化
static void InitSystem(void)
{
	// ポート初期化
	PORTA = PORTB = 0x00;
	LATA = LATB =0x00;
	
	// デバッグLED
	TRISAbits.TRISA0 = 0;
	ANSELAbits.ANSA0 = 0;
	DEBUG_LED = 1;
	
	// COMBO384 F3
	TRISBbits.TRISB12 = 1;
	ANSELBbits.ANSB12 = 0;
	
	// COMBO384 F2
	TRISBbits.TRISB11 = 1;
	
	// COMBO384 F1
	TRISBbits.TRISB10 = 1;
	
	// COMBO384 F0
	TRISBbits.TRISB9 = 1;
	
	// COMBO384 DSDON
	TRISBbits.TRISB8 = 1;
	
	// COMBO384 DSD64
	TRISBbits.TRISB7 = 1;
	CNPUBbits.CNPUB7 = 1; // open drain
	
	// COMBO384 MUTE
	TRISBbits.TRISB5 = 1;
	
	// SSP(I2S)
	ANSELBbits.ANSB13 = 0;
	ANSELBbits.ANSB14 = 0;
	ANSELBbits.ANSB15 = 0;

	SYSKEY = 0xAA996655;
	SYSKEY = 0x556699AA;
	SDI2Rbits.SDI2R = 0b0011; // RPB13
	SS2Rbits.SS2R = 0b0001; // RPB14
	SYSKEY = 0x00000000;

	SPI2CON = 0;
	SPI2CON2 = 0;
	SPI2STATCLR = 0;
	
	SPI2CONbits.FRMEN = 1;
	SPI2CONbits.FRMSYNC = 1;
	SPI2CONbits.FRMPOL = 0;
	SPI2CONbits.FRMSYPW = 0;
	SPI2CONbits.FRMCNT = 0b010;
	SPI2CONbits.SPIFE = 0;
	SPI2CONbits.ENHBUF = 1;
	SPI2CONbits.DISSDO = 1;
	SPI2CONbits.MODE32 = 1;
	SPI2CONbits.MODE16 = 0;
	SPI2CONbits.CKE = 0;
	SPI2CONbits.SSEN = 1;
	SPI2CONbits.CKP = 1;
	SPI2CONbits.MSTEN = 0;
	SPI2CONbits.DISSDI = 0;
	
	SPI2CON2bits.SPISGNEXT = 0;
	SPI2CON2bits.FRMERREN = 1;
	SPI2CON2bits.SPIROVEN = 1;
	SPI2CON2bits.IGNROV = 1;
	SPI2CON2bits.AUDEN = 1;
	SPI2CON2bits.AUDMONO = 0;
	SPI2CON2bits.AUDMOD = 0b00;
	
	// I2C
	ANSELBbits.ANSB2 = 0;
	ANSELBbits.ANSB3 = 0;

	I2C2CONbits.STREN = 1;
	I2C2ADD = I2C_SLAVE;
	IPC9bits.I2C2IP = 6;
	
	// Timer1
	T1CONbits.TCKPS = 0b000; // 1:1 prescale value
	T1CONbits.TSYNC = 0; //External clock input is not synchronized
	T1CONbits.TCS = 1; //  External clock from TxCK pin

	// Timer2
	T2CONbits.TCKPS = 0b000; // 1:1 prescale value
	T2CONbits.TCS = 0; // Internal peripheral clock (0.5us at 2MHz)
	PR2 = 2000; // 1ms

	// 2MHz駆動
	Clock(FALSE);

	// 割り込み許可
	INTCONbits.MVEC = 1; // マルチベクタ割り込み
	__builtin_enable_interrupts();	
}

// ●実行ループ
static void Run(void)
{
	s_nDetected = 0;
	static volatile UINT32 s_u16Bit;
	static volatile UINT32 s_u24Bit;
	static volatile UINT32 s_u32Bit;

	EnableWdt();

	// I2C受付開始
	I2C2CONbits.ON = 1;
	I2C2RCV = I2C2RCV;
	IFS1bits.I2C2SIF = 0;
	IEC1bits.I2C2SIE = 1;
	
	// 電源断まで無限ループ
	int nType = -1;
	UINT32 data;
	for(;;)
	{
		ClrWdt();

		if (COMBO384_MUTE)
		{
			// サンプルレート変更中の場合
			s_u16Bit = 0;
			s_u24Bit = 0;
			s_u32Bit = 0;
			s_nDetected = 0;
			
			while(COMBO384_MUTE)
				ClrWdt();

			data = SPI2BUF;
			continue;
		}
		if (!COMBO384_DSDON)
		{
			// I2Sの場合
			if (nType != 0) // 0:I2S
			{
				T1CONbits.ON = 0;
				T2CONbits.ON = 0;
				SPI2CONbits.ON = 1;
				data = SPI2BUF;
				
				s_u16Bit = 0;
				s_u24Bit = 0;
				s_u32Bit = 0;
				s_nDetected = 0;

				nType = 0;
				continue;
			}
			
			// 受信データの解析
			if (SPI2STATbits.SPIRBE)
				continue;

			data = SPI2BUF;
			if (data == 0x00000000 || data == 0xFFFFFFFF)
			{
				// 判別不能
				continue;
			}

			if ((data & 0x0000FFFF) == 0)
			{
				// 16ビット
				if (s_u16Bit < CHECK_COUNT_MAX) ++s_u16Bit;
				if (s_u24Bit > 0) --s_u24Bit;
				if (s_u32Bit > 0) --s_u32Bit;
			}
			else if ((data & 0x000000FF) == 0)
			{
				// 24ビット
				if (s_u24Bit < CHECK_COUNT_MAX) ++s_u24Bit;
				if (s_u16Bit > 0) --s_u16Bit;
				if (s_u32Bit > 0) --s_u32Bit;
			}
			else
			{
				// 32ビット
				if (s_u32Bit < CHECK_COUNT_MAX) ++s_u32Bit;
				if (s_u16Bit > 0) --s_u16Bit;
				if (s_u24Bit > 0) --s_u24Bit;
			}

			UINT8 nTemp = 0;
			if (s_u16Bit > CHECK_COUNT_THR) nTemp = 16;
			else if (s_u24Bit > CHECK_COUNT_THR) nTemp = 24;
			else if (s_u32Bit > CHECK_COUNT_THR) nTemp = 32;
			s_nDetected = nTemp;
		}
		else
		{
			// DSDの場合
			if (nType != 1) // 1:DSD
			{
				T1CONbits.ON = 1;
				T2CONbits.ON = 1;
				SPI2CONbits.ON = 0;
				s_nDetected = 0;

				nType = 1;
				continue;
			}
			
			// Timer1でDCLKを非同期カウントする
			TMR1 = 0;
			TMR2 = 0;
			IFS0bits.T2IF = 0;
			while(!IFS0bits.T2IF);
			UINT32 nKHz = TMR1;
			
			// テーブルからサンプリング周波数を探す
			int nSample = DSD_UNKNOWN;
			const SDsdSample* pEntry = s_DsdSam;
			do {
				if (pEntry->nMin <= nKHz && nKHz <= pEntry->nMax)
				{
					nSample = pEntry->nSam;
					break;
				}
				pEntry++;
			}
			while(pEntry->nSam != DSD_UNKNOWN);
			
			s_nDetected = nSample;
		}
	}
}

// ●クロック切り替え
static void Clock(BOOL bFast)
{
	UINT32 uHz;
	UINT32 uOSCCON;
	
	if (bFast)
	{
		// SYSCLK=48MHz駆動
		uOSCCON = OSCCON & 0xFFFFF8FF;
		uOSCCON = uOSCCON | (0b001 << 8); // Internal Fast RC Oscillator with PLL module via Postscaler (FRCPLL)
		uHz = 48000000UL;
	}
	else
	{
		// SYSCLK=2MHz駆動
		uOSCCON = OSCCON & 0xF8FFF8FF;
		uOSCCON = uOSCCON | (0b111 << 8); // Internal Fast RC (FRC) Oscillator divided by FRCDIV<2:0> bits
		uOSCCON = uOSCCON | (0b010 << 24); // FRC divided by 4
		uHz = 2000000UL;	
	}

	__builtin_disable_interrupts();

	SYSKEY = 0xAA996655;
	SYSKEY = 0x556699AA;
	OSCCON = uOSCCON;
	OSCCON = uOSCCON | 1; // OSWEN
	SYSKEY = 0x00000000;
	
	while(OSCCONbits.OSWEN);
	
	SYSTEMConfigPerformance(uHz);	

	__builtin_enable_interrupts();
}

// ●割り込み
void __attribute__((vector(_I2C_2_VECTOR), interrupt(IPL6AUTO), nomips16))
I2CISR(void)
{
	if (IFS1bits.I2C2SIF)
	{
		BYTE data;
	
		switch(I2C2STAT & 0b000100110)
		{
		case 0b000000110: // 送信(1Byte目)
			// データ(D/A=0) マスターへ送信(R/W=1) 受信完了(RBF=1)
			data = I2C2RCV;

			// 入力状態を返す
			data = 0;
			if (COMBO384_F3) data |= COMBO384_F3_MASK;
			if (COMBO384_F2) data |= COMBO384_F2_MASK;
			if (COMBO384_F1) data |= COMBO384_F1_MASK;
			if (COMBO384_F0) data |= COMBO384_F0_MASK;
			if (COMBO384_DSDON) data |= COMBO384_DSDON_MASK;
			if (COMBO384_DSD64) data |= COMBO384_DSD64_MASK;
			if (COMBO384_MUTE) data |= COMBO384_MUTE_MASK;

			I2C2TRN = data;
			Nop(); // SCLRELをセットするまでの間に遅延時間を設ける必要がある
			
			I2C2CONbits.SCLREL = 1;
			break;

		case 0b000100100: // 送信(2Byte目以降)
			// データ(D/A=1) マスターへ送信(R/W=1) 受信完了(RBF=0)
			if (I2C2STATbits.ACKSTAT == 0)
			{
				// I2Sの場合は量子化ビット数、DSDの場合はサンプリング周波数を返す
				I2C2TRN = s_nDetected;
				Nop(); // SCLRELをセットするまでの間に遅延時間を設ける必要がある
			}
			I2C2CONbits.SCLREL = 1;
			break;

		case 0b000100010: // 受信
			// データ(D/A=1) マスターから受信(R/W=0) 受信完了(RBF=1)
			data = I2C2RCV;

			// コマンド受信
			// 何もしない

			I2C2CONbits.SCLREL = 1;
			break;

		default:
			data = I2C2RCV;
			I2C2CONbits.SCLREL = 1;
			break;
		}

		IFS1bits.I2C2SIF = 0;
	}
}
