/****************************************************************************

 ■HiResAmp 入力監視(InputSection)
		Copyright(C) 2020 電子うさぎ https://電子うさぎ.com/

****************************************************************************/

// 旧来のplibから「SYSTEMConfigPerformance」に関する一部を持ってきた

#include <xc.h>
#include <proc/p32mx120f032b.h>


/********************************************************************* 
 * Function:        INTDisableInterrupts() 
 * 
 * PreCondition:    None 
 * 
 * Input:           None 
 * 
 * Output:          None 
 * 
 * Side Effects:    Interrupts are disabled 
 * 
 * Overview:        Interrupts are disabled by clearing the IE bit 
 *                  in the status register 
 ********************************************************************/ 
extern inline unsigned int __attribute__((always_inline)) INTDisableInterrupts(void)  
{ 
    unsigned int status = 0; 
     
    asm volatile("di    %0" : "=r"(status)); 
 
    return status; 
}

/********************************************************************* 
 * Function:        INTRestoreInterrupts(unsigned int status) 
 * 
 * PreCondition:    None 
 * 
 * Input:           value of the status registor 
 * 
 * Output:           
 * 
 * Side Effects:    Interrupts are restored to previous state 
 * 
 * Overview:        Interrupts are enabled by setting the IE bit 
 *                  in the status register 
 ********************************************************************/ 
extern inline void __attribute__((always_inline)) INTRestoreInterrupts(unsigned int status) 
{ 
    if(status & 0x00000001) {
		asm volatile("ei");
	}
    else {
		asm volatile("di");
	}
}

#define mBMXDisableDRMWaitState()   (BMXCONCLR = _BMXCON_BMXWSDRM_MASK)

// ※本プロジェクトではDMAは使わない
#define DmaSuspend()				0
#define DmaResume(x)

#ifdef _DMAC
    #define mSYSTEMLock(intStat, dmaSusp)   do{SYSKEY = 0x33333333; DmaResume(dmaSusp); INTRestoreInterrupts(intStat);}while(0)
#else
    #define mSYSTEMLock(intStat, dmaSusp)   do{SYSKEY = 0x33333333; INTRestoreInterrupts(intStat);}while(0)
#endif // _DMAC

#ifdef _DMAC
    #define mSYSTEMUnlock(intStat, dmaSusp) do{intStat=INTDisableInterrupts(); dmaSusp=DmaSuspend(); \
                        SYSKEY = 0, SYSKEY = 0xAA996655, SYSKEY = 0x556699AA;}while(0)
#else
    #define mSYSTEMUnlock(intStat, dmaSusp) do{intStat=INTDisableInterrupts(); \
                        SYSKEY = 0, SYSKEY = 0xAA996655, SYSKEY = 0x556699AA;}while(0)
#endif  // _DMAC


/*********************************************************************
 * Function:    OSCSetPBDIV(unsigned int oscPbDiv)
 *
 * Description: Configures peripheral bus divisor
 *
 * PreCondition: oscPbDiv one of the OSC_PB_DIV_X valid values
 *
 * Inputs:    oscPbDiv - desired PB divider
 *
 * Output:      None
 *
 * Example:   OSCSetPBDIV(  OSC_PB_DIV_8 )
 *
 ********************************************************************/
extern inline void __attribute__((always_inline)) OSCSetPBDIV(unsigned int oscPbDiv)
{
    unsigned int dma_status;
    unsigned int int_status;
    __OSCCONbits_t oscBits;

    mSYSTEMUnlock(int_status, dma_status);

    oscBits.w=OSCCON;       // read to be in sync. flush any pending write
    oscBits.PBDIV=0;
    oscBits.w|=oscPbDiv;
    OSCCON=oscBits.w;       // write back
    oscBits.w=OSCCON;       // make sure the write occurred before returning from this function

    mSYSTEMLock(int_status, dma_status);
}

// CPU Peripheral Bus divisor values - values are mutually exclusive
#define OSC_PB_DIV_8    (3 << _OSCCON_PBDIV_POSITION)
#define OSC_PB_DIV_4    (2 << _OSCCON_PBDIV_POSITION)
#define OSC_PB_DIV_2    (1 << _OSCCON_PBDIV_POSITION)
#define OSC_PB_DIV_1    (0 << _OSCCON_PBDIV_POSITION)


#define PB_BUS_MAX_FREQ_HZ      80000000 //Max Peripheral bus speed

/*********************************************************************
 Function:
    unsigned int SYSTEMConfigPB(unsigned int sys_clock)

 Description:
    The function sets the PB divider to the optimum value.

 PreCondition:
    None

 Parameters:
    sys_clock - system clock in Hz

 Returns:
    the PB clock frequency in Hz

 Side Effects:
    The PB clock may be changed

 Overview:
    The function sets the PB divider to the optimum value.

 Remarks:
    The interrupts are disabled briefly, the DMA is suspended and the system is unlocked while performing the operation.
    Upon return the previous status of the interrupts and the DMA are restored. The system is re-locked.

 Example:
    <code>
    SYSTEMConfigPB(72000000);
    </code>
 ********************************************************************/
extern inline unsigned int __attribute__((always_inline)) SYSTEMConfigPB(unsigned int sys_clock)
{
    unsigned int pb_div;
    unsigned int pb_clock;

    pb_clock = sys_clock;

    if(sys_clock > PB_BUS_MAX_FREQ_HZ)
    {
        pb_div=OSC_PB_DIV_2;
        pb_clock >>= 1;
    }
    else
    {
        pb_div=OSC_PB_DIV_1;
    }

    OSCSetPBDIV(pb_div);

    return pb_clock;
}

/*********************************************************************
  Function:
        unsigned int SYSTEMConfigWaitStatesAndPB(unsigned int sys_clock)

  Description:
    The function sets the PB divider and the Flash Wait states to the optimum value.

  PreCondition:
    None

  Parameters:
    sys_clock - system clock in Hz

  Returns:
    the PB clock frequency in Hz

  Side Effects:
    The PB clock and wait states may be changed

  Remarks:
    The interrupts are disabled briefly, the DMA is suspended and the system is unlocked while performing the operation.
    Upon return the previous status of the interrupts and the DMA are restored. The system is re-locked.

  Example:
    <code>
    SYSTEMConfigWaitStatesAndPB(72000000);
    </code>
 ********************************************************************/
extern inline unsigned int __attribute__((always_inline)) SYSTEMConfigWaitStatesAndPB(unsigned int sys_clock)
{
#ifdef _PCACHE
    unsigned int wait_states;
#endif
    unsigned int pb_clock;
    unsigned int int_status;

    pb_clock = SYSTEMConfigPB(sys_clock);

    // set the flash wait states based on 1 wait state
    // for every 20 MHz
#ifdef _PCACHE
    wait_states = 0;

    while(sys_clock > FLASH_SPEED_HZ)
    {
        wait_states++;
        sys_clock -= FLASH_SPEED_HZ;
    }

    int_status=INTDisableInterrupts();
    mCheConfigure(wait_states);
    INTRestoreInterrupts(int_status);

#endif

    return pb_clock;
}

/*********************************************************************
  Function:
    unsigned int SYSTEMConfigPerformance(unsigned int sys_clock)

  Description:
    The function sets the PB divider, the Flash Wait states and the DRM wait states to the optimum value.
    It also enables the cacheability for the K0 segment.

  PreCondition:
    None

  Parameters:
    sys_clock - system clock in Hz

  Output:
    the PB clock frequency in Hz

  Side Effects:
    Sets the PB and Flash Wait states

  Remarks:
    The interrupts are disabled briefly, the DMA is suspended and the system is unlocked while performing the operation.
    Upon return the previous status of the interrupts and the DMA are restored. The system is re-locked.

  Example:
    <code>
    SYSTEMConfigPerformance(72000000);
    </code>
 ********************************************************************/
extern unsigned int SYSTEMConfigPerformance(unsigned int sys_clock)
{
    // set up the wait states
    unsigned int pb_clk;
#ifdef _PCACHE
    unsigned int cache_status;
#endif
    unsigned int int_status;

    pb_clk = SYSTEMConfigWaitStatesAndPB(sys_clock);

    int_status=INTDisableInterrupts();

    mBMXDisableDRMWaitState();

#ifdef _PCACHE
    cache_status = mCheGetCon();
    cache_status |= CHE_CONF_PF_ALL;
    mCheConfigure(cache_status);
    CheKseg0CacheOn();
#endif

    INTRestoreInterrupts(int_status);

    return pb_clk;
}
