/****************************************************************************

 ■HiResAmp バイアス制御(PowerAmp)
		Copyright(C) 2020 電子うさぎ https://電子うさぎ.com/

****************************************************************************/

// PIC16F1705 Configuration Bit Settings

// CONFIG1
#pragma config FOSC = INTOSC    // Oscillator Selection Bits (INTOSC oscillator: I/O function on CLKIN pin)
#pragma config WDTE = SWDTEN    // Watchdog Timer Enable (WDT controlled by the SWDTEN bit in the WDTCON register)
#pragma config PWRTE = OFF      // Power-up Timer Enable (PWRT disabled)
#pragma config MCLRE = ON       // MCLR Pin Function Select (MCLR/VPP pin function is MCLR)
#pragma config CP = OFF         // Flash Program Memory Code Protection (Program memory code protection is disabled)
#pragma config BOREN = ON       // Brown-out Reset Enable (Brown-out Reset enabled)
#pragma config CLKOUTEN = OFF   // Clock Out Enable (CLKOUT function is disabled. I/O or oscillator function on the CLKOUT pin)
#pragma config IESO = ON        // Internal/External Switchover Mode (Internal/External Switchover Mode is enabled)
#pragma config FCMEN = ON       // Fail-Safe Clock Monitor Enable (Fail-Safe Clock Monitor is enabled)

// CONFIG2
#pragma config WRT = OFF        // Flash Memory Self-Write Protection (Write protection off)
#pragma config PPS1WAY = ON     // Peripheral Pin Select one-way control (The PPSLOCK bit cannot be cleared once it is set by software)
#pragma config ZCDDIS = ON      // Zero-cross detect disable (Zero-cross detect circuit is disabled at POR)
#pragma config PLLEN = OFF      // Phase Lock Loop enable (4x PLL is enabled when software sets the SPLLEN bit)
#pragma config STVREN = ON      // Stack Overflow/Underflow Reset Enable (Stack Overflow or Underflow will cause a Reset)
#pragma config BORV = LO        // Brown-out Reset Voltage Selection (Brown-out Reset Voltage (Vbor), low trip point selected.)
#pragma config LPBOR = OFF      // Low-Power Brown Out Reset (Low-Power BOR is disabled)
#pragma config LVP = ON         // Low-Voltage Programming Enable (Low-voltage programming enabled)

#include <xc.h>

//=================================================================

// I2C Slave アドレス
//#define I2C_SLAVE	0x22	// R+
//#define I2C_SLAVE	0x23	// R-
//#define I2C_SLAVE	0x20	// L+
#define I2C_SLAVE	0x21	// L-

// 設定JMPピン
#define JMP_VR_SET			RA1

// バイアス制御設定
#define RAISE_INTERVAL		50		// バイアスを上げる間隔(ms)
#define UPPER_LIMIT			600		// 平均電流上限(mA)
#define LOWER_INTERVAL_MIN	100		// バイアスを下げる最小間隔(ms)
#define LOWER_INTERVAL_MAX	1200	// バイアスを下げる最大間隔(ms)
#define LOWER_RATE			10		// バイアスを下げる間隔短縮率(%)
#define LOWER_MARGIN		9		// スタンバイマージン(mA)

// ステート値
#define STATE_INIT			0		// 初期化中
#define STATE_NORMAL		1		// 通常状態
#define STATE_STANDBY		2		// スタンバイ状態
#define STATE_TEST			3		// テストモード
#define STATE_SHUTDOWN		5		// シャットダウン(共通)
#define STATE_WARNING		7		// DAC0なのに下限に達しない
#define STATE_ERROR1		8		// 平均電流上限超過
#define STATE_ERROR2		9		// WDTタイムアウト発生

// コマンド
#define CMD_SHUTDOWN		99		// シャットダウン(共通)

//---------------------------------------------------------------

typedef enum _BOOL { FALSE = 0, TRUE } BOOL;
typedef unsigned char UINT8;
typedef unsigned short int UINT16;

//---------------------------------------------------------------

static void InitSystem(void);
static void Run(void);
static BOOL WaitDetect(UINT16 nTimeout);
static void DelayMs(UINT16 nCount);

#define MAKE_WORD(h, l)		(((UINT16)(h) << 8) | (l))

//---------------------------------------------------------------

static UINT8 s_nState;
static UINT16 s_nAvgCurrent;
static UINT16 s_nLowerLimit;

static UINT16 s_nLowerInterval;
static volatile UINT8 s_nExit;

static UINT16 s_nSendData;
static UINT16 s_nWaitCount;

//=================================================================

// ●メインエントリ
void main(void)
{
	InitSystem();
	Run();
}

// ●システム初期化
static void InitSystem(void)
{
	// クロック設定
	OSCCONbits.IRCF = 0b1110; // 8MHz HF 

	// ポート初期化
	PORTA = PORTC = 0x00;
	LATA = LATC = 0x00;
	WPUA = WPUC = 0x00;
	nWPUEN = 0;
	
    // 空き(RC1)
	TRISCbits.TRISC1 = 0;
    ANSELCbits.ANSC1 = 0;
	
    // JMP(VR_SET)(RA1)
	WPUAbits.WPUA1 = 1;
	TRISAbits.TRISA1 = 1;
    ANSELAbits.ANSA1 = 0;
	
	// FVR
	FVRCONbits.CDAFVR = 0b10; // 2x VFVR (2.048V))
	FVRCONbits.FVREN = 1;
	while(!FVRCONbits.FVRRDY);
	
	// DAC
	DAC1CON0bits.DAC1OE2 = 1;
	DAC1CON0bits.DAC1PSS = 0b10; // FVR Buffer2 output
	DAC1CON0bits.DAC1NSS = 0b00; // VSS
	DAC1CON0bits.DAC1EN = 1;
	DAC1CON1bits.DAC1R = 0;
	
	// OPA
	OPA2CONbits.OPA2PCH = 0b00; // Non-inverting input connects to OPAxIN+ pin
	OPA2CONbits.OPA2UG = 0; // Inverting input is connected to the OPAxIN- pin
	OPA2CONbits.OPA2SP = 1; // Op amp operates in high GBWP mode
	OPA2CONbits.OPA2EN = 1;
	
	// Comparer1
	CM1CON0bits.C1HYS = 0;
	CM1CON0bits.C1SP = 1;
	CM1CON0bits.C1ZLF = 1; // Comparator output is filtered
	CM1CON0bits.C1POL = 1; //  Comparator output is inverted
	CM1CON1bits.C1NCH = 0b010; // CxVN connects to CxIN2- pin
	CM1CON1bits.C1PCH = 0b000; // CxVP connects to CxIN+ pin
	CM1CON1bits.C1INTP = 1;
	CM1CON1bits.C1INTN = 0;
	CM1CON0bits.C1ON = 1;
	
	// Comparer2
	CM2CON0bits.C2HYS = 0;
	CM2CON0bits.C2SP = 1;
	CM2CON0bits.C2ZLF = 1; // Comparator output is filtered
	CM2CON0bits.C2POL = 1; //  Comparator output is inverted
	CM2CON1bits.C2NCH = 0b010; // CxVN connects to CxIN2- pin
	CM2CON1bits.C2PCH = 0b000; // CxVP connects to CxIN+ pin
	CM2CON1bits.C2INTP = 1;
	CM2CON1bits.C2INTN = 0;
	CM2CON0bits.C2ON = 1;
	
	// ADC
	ADCON1bits.ADFM = 1; // Right justified
	ADCON1bits.ADCS = 0x001; // FOSC/8(TAD=1us at Fosc=8MHz)
	ADCON1bits.ADNREF = 0; // VREF- is connected to VSS
	ADCON1bits.ADPREF = 0b00; // VREF+ is connected to VDD
	ADCON2bits.TRIGSEL = 0b0100; // Timer1 – T1_overflow
	ADCON0bits.ADON = 1;
	
	// Timer1(ADCトリガー 0.5us at 8MHz FOSC)
	T1CONbits.TMR1CS = 0b00; // Timer1 clock source is instruction clock (FOSC/4)
	T1CONbits.T1CKPS = 0b10; // 1:4 Prescale value
	
	// Timer2(Deray用 0.5us at 8MHz FOSC)
	T2CONbits.T2CKPS = 0b00; // Prescaler is 1
	T2CONbits.T2OUTPS = 0b1001; // 1:10 Postscaler
	PR2 = 200; // 1ms
	
	// I2C(Slave mode) (SCL:RC0→RA4, SDA:RC1→RA5)
	ANSELAbits.ANSA4 = 0;
	SSPCLKPPS = 0b00100; // Peripheral input is RA4
	SSPDATPPS = 0b00101; // Peripheral input is RA5
	RA4PPS = 0b10000; // RA4 source is SCK/SCL
	RA5PPS = 0b10001; // RA5 source is SDA
	
	SSP1ADD = I2C_SLAVE << 1;
	SSP1MSK = 0b11111110;
	SSP1STAT= 0b00000000;

	SSP1CON1bits.SSPM = 0b0110; // I2C Slave mode, 7-bit address 
	SSP1CON2bits.SEN = 1; // Clock stretching is enabled
	SSP1CON1bits.CKP = 1;
	
	// WDT
	WDTCONbits.WDTPS = 0b01101; // Interval 8s nominal
	
	// 割込み許可
	ADIE = 1;
	SSP1IE = 1;
	PEIE = 1;
	GIE = 1;
}

// ●実行ループ
static void Run(void)
{
#ifndef _DEBUG
	SWDTEN = 1;
#endif
	// アプリ初期化
	s_nState = STATE_INIT;
	s_nAvgCurrent = 0;
	s_nLowerLimit = 0;
	s_nLowerInterval = LOWER_INTERVAL_MAX;
	s_nExit = 0;
	
	// I2C受付開始
	SSP1CON1bits.SSPEN =1;

	// 電流値取得開始
	ADCON0bits.CHS = 0b00110; // AN6(8pin)
	TMR1ON = 1;

#ifndef _DEBUG
	// 起動時にジャンパ挿入でバイアス無しモード(DACOUT=0V)
	if (!JMP_VR_SET)
	{
		s_nState = STATE_TEST;

		while(!JMP_VR_SET) // ジャンパを抜くまで
			CLRWDT();

		s_nState = STATE_INIT;
	}
#endif
  	// WDT復帰チェック
	if (PCONbits.nRWDT != 0)
	{
		// 通常動作開始
		s_nState = STATE_NORMAL;
	}
	else
	{
		// タイムアウト発生
		s_nState = STATE_ERROR2;
	}

	// 安定待ち
	DelayMs(500);
	C1IF = 0;
	C2IF = 0;

	// シャットダウンまで無限ループ
	while(s_nExit == 0)
	{
		CLRWDT();
		
		// エラーチェック
		if (s_nState >= STATE_ERROR1)
		{
			// エラー発生時はシャットダウンまで何もしない
			DAC1CON1bits.DAC1R = 0x00;
			continue;
		}

#ifndef _DEBUG
		// ジャンパ挿入でVI変換ゲイン調整モード(DACOUT=3.3V)
		if (!JMP_VR_SET)
		{
			s_nState = STATE_TEST;

			DAC1CON1bits.DAC1R = 0xFF;
			while(!JMP_VR_SET) // ジャンパを抜くまで
				CLRWDT();

			// バイアス無しにして再開
			DAC1CON1bits.DAC1R = 0x00;
			C1IF = 0;
			C2IF = 0;

			s_nLowerInterval = LOWER_INTERVAL_MAX;
			s_nState = STATE_NORMAL;
			continue;
		}
#endif
		// バイアス制御
		if (WaitDetect(s_nLowerInterval))
		{
			// 指示値より低下を検知 →バイアスを上げる
			if (DAC1CON1bits.DAC1R < 0xFF)
				++DAC1CON1bits.DAC1R; // 出力コンデンサがあるためすぐには上がらない
			
			s_nLowerInterval = LOWER_INTERVAL_MAX;
			s_nState = STATE_NORMAL;

			DelayMs(RAISE_INTERVAL);

			// 出力が上がってからフラグクリア
			C1IF = 0;
			C2IF = 0;
			DelayMs(1);
		}
		else
		{
			// 指示値より高い可能性 →バイアスを下げる
			if (s_nAvgCurrent >= (s_nLowerLimit + LOWER_MARGIN))
			{
				// バイアスを下げる
				if (DAC1CON1bits.DAC1R > 0x00)
				{
					--DAC1CON1bits.DAC1R;

					// 下げる間隔を短くしていく
					if (s_nLowerInterval > LOWER_INTERVAL_MIN)
					{
						s_nLowerInterval -= (s_nLowerInterval * LOWER_RATE) / 100;
						if (s_nLowerInterval < LOWER_INTERVAL_MIN)
							s_nLowerInterval = LOWER_INTERVAL_MIN;
					}
				}
				else
				{
					// DAC0出力にしたのに下限に達しない
					s_nLowerInterval = LOWER_INTERVAL_MAX;
					s_nState = STATE_WARNING;
				}
			}
			else
			{
				// これ以上は下げない→スタンバイ状態
				s_nLowerInterval = LOWER_INTERVAL_MAX;
				s_nState = STATE_STANDBY;
			}
		}
	}
	
	// シャットダウン(電源断まで無限ループ	)
	DAC1CON1bits.DAC1R = 0x00;
	TMR1ON = 0;
	ADCON0bits.ADON = 0;

	s_nState = s_nExit;

	for(;;) CLRWDT();
}

// ●下限超過検出待ち
static BOOL WaitDetect(UINT16 nTimeout)
{
	BOOL bDetect = FALSE;
	
	if (nTimeout > 0)
	{
		s_nWaitCount = nTimeout;
		TMR2 = 0;

		TMR2IE = 1;
		TMR2ON = 1;
		while(TMR2ON)
		{
			if (C1IF || C2IF || CM1CON0bits.C1OUT || CM2CON0bits.C2OUT)
			{
				bDetect = TRUE;
				break;
			}
		}
		TMR2ON = 0;
		TMR2IE = 0;
		TMR2IF = 0;
	}
	
	return bDetect;
}

// ●ディレイ
static void DelayMs(UINT16 nCount)
{
	if (nCount > 0)
	{
		s_nWaitCount = nCount;
		TMR2 = 0;

		TMR2IE = 1;
		TMR2ON = 1;
		while(TMR2ON);
	}
}

//-----------------------------------------------------------------

// ●割込み
void __interrupt() Interrupt(void)
{
	// ADC
	if (ADIF)
	{
		// 16回分(393ms✕8(PNP/NPN) 約3秒間)を記録する
		if (ADCON0bits.CHS != 0b00110)
		{
			// 過去1秒間の平均電流の取得
			static UINT16 nCurrent[16] = { 0 };
			static UINT8 nCount = 0xFF;

			UINT16 nVal = (ADRESH << 8) | ADRESL;
			int i = 0;
			if (nCount != 0xFF)
			{
				nCurrent[nCount] = nVal;
				if (++nCount >= 16) nCount = 0;

				nVal = 0;
				do nVal += nCurrent[i];
				while (++i < 16);
			}
			else
			{
				// 一番最初は同じ値を詰めておく
				do nCurrent[i] = nVal;
				while (++i < 16);
				nVal <<= 4; // x16

				nCount = 1;
			}
			// 平均を求めて電流(mA)へ変換する
			// mA = (((3300 * [DAC Values] / 16) / 1023) / 8.393) / 0.47
			//     ≒ (3300 * [DAC Values] / 16) / 4035
			s_nAvgCurrent = (3300UL * nVal / 16) / 4035UL;
			
			// 平均電流リミットチェック
			if (s_nAvgCurrent >= UPPER_LIMIT)
				s_nExit = STATE_ERROR1; // シャットダウンする
		}
		else
		{
			// バイアス設定値の読み取り
			UINT16 nVal = (ADRESH << 8) | ADRESL;
			s_nLowerLimit = (3300UL * nVal) / 4035UL; 
		}
		
		// 入力チャンネルを順番に切り替える
		switch(ADCON0bits.CHS)
		{
		case 0b00110: // AN6(8pin)
			ADCON0bits.CHS = 0b00000; // AN0(13pin)
			break;
		case 0b00000: // AN0(13pin)
			ADCON0bits.CHS = 0b00100; // AN4(10pin)
			break;
		case 0b00100: // AN4(10pin)
			ADCON0bits.CHS = 0b00110; // AN6(8pin)
			break;
		}
		ADIF = 0;
	}

	// I2Cスレーブ通信
	if (SSP1IF)
	{
		static volatile UINT8 s_I2CCount = 0;
		UINT8 temp;

		// D/A,R/W,BFビットを調査
		switch (SSP1STAT & 0b00100101)
		{
		case 0b00000101:
			// アドレス(D/A=0) マスターへ送信(R/W=1) データあり(BF=1)
			temp = SSP1BUF; // 空読み
			SSP1BUF = s_nState; // 1バイト目(ステータス)
			CKP = 1;
			s_I2CCount = 1;
			break;

		case 0b00100100:
			// アドレス(D/A=1) マスターへ送信(R/W=1) データなし(BF=0)
			switch (s_I2CCount)
			{
			case 1:
				// 2バイト目(DAC値)
				SSP1BUF = DAC1CON1bits.DAC1R;
				break;
			case 2:
				// 3バイト目(現在電流(mA)) LOWORD
				s_nSendData = s_nAvgCurrent;
				SSP1BUF = (UINT8)s_nSendData; 
				break;
			case 3:
				// 4バイト目(現在電流(mA)) HIWORD
				SSP1BUF = (UINT8)(s_nSendData >> 8); 
				break;
			case 4:
				// 5バイト目(設定電流(mA)) LOWORD
				s_nSendData = s_nLowerLimit;
				SSP1BUF = (UINT8)s_nSendData; 
				break;
			case 5:
				// 6バイト目(設定電流(mA)) HIWORD
				SSP1BUF = (UINT8)(s_nSendData >> 8); 
				break;
			default:
				SSP1BUF = 0;
				break;
			}
			CKP = 1;
			s_I2CCount++;
			break;

		case 0b00000001:
			// データ(D/A=0) マスターから受信(R/W=0) データあり(BF=1)
			temp = SSP1BUF; // 空読み
			s_I2CCount = 0;
			CKP = 1;
			break;
			
		case 0b00100001:
			// データ(D/A=1) マスターから受信(R/W=0) データあり(BF=1)
			temp = SSP1BUF;
			
			// コマンド処理
			if (temp == CMD_SHUTDOWN)
				s_nExit = STATE_SHUTDOWN;

			CKP = 1;
			break;

		default:
			temp = SSP1BUF;
			CKP = 1;
			break;
		}
		SSP1IF = 0;
	}	

	// Timer2(Delay用)
	if (TMR2IF)
	{
		if (--s_nWaitCount == 0)
		{
			TMR2ON = 0;
			TMR2IE = 0;
		}
		TMR2IF = 0;
	}
}
